--
-- LLRNet - network part of LLR
--
-- (C) 2004-2005 Vincent Penne
-- Changes 2010, Karsten Bonath
--
-- Released under GNU LIBRARY GENERAL PUBLIC LICENSE
-- (See file LICENSE that must be included with this software)
--
--

dofile(EXEDIR.."init.lua")

if not dolib("basic") then
   return
end

if not dolib("client") then
   return
end

if not dolib("client_server") then
   return
end

server = "localhost"
port = 7000
username = "nobody"
password = "llrnet"
workFile = "workfile.txt"
tosendFile = "tosend.txt"
WUCacheMaxSize = 100
tosend = { }
asynchronous = 1

---------------------------------------------------------------------
-- asynchronous mode :
-- this function dialogs with the server to send latest results and 
-- grabs new work units
-- this function is called regularly in a separate thread
---------------------------------------------------------------------
function DialogWithServer()

   if WUCacheSize == 0 then
     return
   end

   if not username or username == "nobody" then
      return
   end

   SendAllResults()
   local t, k, n, more
   local changed

   -- first check if we have some job to work on
   t, k, n, more = ReadWorkfile()
   more = more or { }

   -- if no unfinished job, then ask a new job to the server
   if not t or not k or not n then
      t, k, n = GetPair()
      if t and k and n then
	 print(format("  Fetching WU #1/%d: %s %s",WUCacheSize,k,n))
         changed = 1
      else
         return
      end
   end

   if t and k and n and WUCacheSize then
      local i, t, k, n
      local s = getn(more) + 2
      local num = WUCacheSize - s + 1
      for i = s, WUCacheSize, 1 do
	 t, k, n = GetPair()
	 if t and k and n then
            print(format("  Fetching WU #%d/%d: %s %s",i,WUCacheSize,k,n))
  	    tinsert(more, { k = k, n = n })
	    changed = 1
	 else
  	    break
	 end
      end
   end

   if changed then
      WriteWorkfile(t, k, n, more)
   end
   Logout()
end

---------------------------------------------------------------------
-- perform some work
-- returns nil if success, -1 if some error occures
---------------------------------------------------------------------
function Work()
   local t, k, n, more
   local result, residue

   SemaWait(semaphore)

--   if not asynchronous then
--     SendAllResults((sendRetries or 0) + 1)
--   end

   -- first check if we have some job to work on
   t, k, n, more = ReadWorkfile()
   more = more or { }

   if not asynchronous and (not t or not k or not n) then
      if cancelJob then
	 print("No more job to cancel !")
	 SemaSignal(semaphore)
	 return
      end
   end

   if k and n and t then
      if not asynchronous and not cancelJob and WUCacheSize then
	 local i, t, k, n
	 local s = getn(more) + 2
	 local num = WUCacheSize - s + 1
	 if num > 0 then
	    print(format("Fetching %d additional Work Unit(s) ...", num))
	 end
	 for i = s, WUCacheSize, 1 do
	    t, k, n = GetPair()
	    if t and k and n then
	       print(format("  Fetching WU #%d/%d: %s %s",i,WUCacheSize,k,n))
	       tinsert(more, { k = k, n = n })
	    else
	       break
	    end
	 end
      end

      WriteWorkfile(t, k, n, more)
      SemaSignal(semaphore)

      if cancelJob then
	print(format("Cancelling : %s/%s", k, n))
 	result, residue = -2, "CANCEL"
        local tbl = { t = t, k = k, n = n, result = result, residue = residue, date = date("%c") }
        tinsert(tosend,tbl)
        while more[1] do
          local p = more[1]
	  local k, n = p.k, p.n
          print(format("Cancelling : %s/%s", k, n))
          local tbl = { t = t, k = k, n = n, result = result, residue = residue, date = date("%c") }
          tinsert(tosend,tbl)
	  tremove(more, 1)
	  if WriteWorkfile(t, k, n, more) then
             print(format("Remove : (%s) %s/%s", k, n, t))
	     print("Could not write to workfile.txt file !")
	     SemaSignal(semaphore)
	     return -1
	  end
        end
        ClearWorkfile()
      else
	 -- perform prime test !
	 if not asynchronous then
	    Logout() -- logout before performing computation
	 end
--	 result, residue = primeTest(t, format("%s %s", k, n))
	 result, residue = 0, "0"
	 -- check user interruption
      end
      SemaWait(semaphore)
--      if not asynchronous then
	 SendAllResults((sendRetries or 0) + 1)
--      end
      SemaSignal(semaphore)
      return nil
   else
      SemaSignal(semaphore)
      if not asynchronous then
	 print("Nothing to do ! Will ask more work from the server a bit later ...")
      end
      return -1
   end
end

---------------------------------------------------------------------
-- Help function
---------------------------------------------------------------------
function Help()
   print([[
usage : llrnet [options]
options :
 -h : print this message
 -c : cancel current work unit
]])
end

---------------------------------------------------------------------
-- LLR Main program loop
---------------------------------------------------------------------
function LLRMain()
   -- check options
   if not username or username == "nobody" then
      paused = 1
      print("Please enter a username into llr-clientconfig.txt file.")
      return
   end

   -- main loop
   
--   if not asynchronous then
      ReadTosendfile()
--      local unsent = tosend[1]
      SendAllResults()
--      if unsent then
--	 Logout()
--	 return
--      end
--   end

   if Work() then
      if not asynchronous then
         Logout()
      end
      -- sleep one minute if some error occured
      if not asynchronous then
         print("Sleeping ...")
      end
      local i
      sleeping = 1
      if asynchronous then
         sleep(1)
      else
         for i=1,60,1 do
            if not sleeping then
               break
            end
            sleep(1)
         end
         SendAllResults()
      end
   end
      
   Logout()
end

---------------------------------------------------------------------
-- Main Program
---------------------------------------------------------------------
-- read configuration file
dofile("llr-clientconfig.txt")

-- parse command line options
local i, n
n = getn(arg)
for i=1,n,1 do
   if arg[i] == "-c" then
      cancelJob = 1
   elseif arg[i] == "-h" then
      Help()
      return
   else
      print(format("Unknown option '%s'", arg[i]))
      Help()
      return
   end
end

-- put a hard limit to the WUCacheSize
WUCacheSize = min(WUCacheSize or 1, WUCacheMaxSize)

-- create the global file access semaphore
semaphore = SemaCreate(1, 1)
   
if cancelJob then
   asynchronous = nil
end

if asynchronous then
   ReadTosendfile()
   DialogWithServer()
   serverDialogThreadRunning = 1
   create_thread(function()
     while serverDialogThreadRunning ~= -1 do
        -- We retry the server every minute except when we get a new result, then we retry at once
        local i, n
        n = getn(tosend)
        for i=1,60,1 do
           sleep(1)
           if serverDialogThreadRunning == -1 or n ~= getn(tosend) then
              break
           end
        end
        if serverDialogThreadRunning ~= -1 then
           SemaWait(semaphore)
           DialogWithServer()
           SemaSignal(semaphore)
        end
     end
     serverDialogThreadRunning = nil
    end)
end
   
while true do
   LLRMain()

   if paused then
      sleeping = 1
      sleep(1)
      sleeping = nil
   else
      break
   end
end

if asynchronous then
   serverDialogThreadRunning = -1
   while serverDialogThreadRunning do
      sleep(1)
   end
end
